@@ -8,7 +8,7 @@ use Mojo::Util 'steady_time';
 use Sys::Hostname 'hostname';
 use Time::HiRes 'usleep';
 
-our $VERSION = 'v5.0.2';
+our $VERSION = 'v5.0.3';
 
 has dequeue_interval => 0.5;
 has 'sqlite';
@@ -47,18 +47,17 @@ sub dequeue {
 sub enqueue {
   my ($self, $task, $args, $options) = (shift, shift, shift || [], shift || {});
 
-  my $db = $self->sqlite->db;
-  return _enqueue($db, $task, $args, $options) unless my $seq = $options->{sequence};
-
-  my $tx = $db->begin('exclusive');
-  my $prev = $db->query(
-    q{select id from minion_jobs where sequence = ? and next is null limit 1}, $seq)
-    ->hashes->first;
-  unshift @{$options->{parents}}, $prev->{id} if $prev;
-  my $id = _enqueue($db, $task, $args, $options);
-  $db->query(q{update minion_jobs set next = ? where id = ?}, $id, $prev->{id}) if $prev;
-  $tx->commit;
-  return $id;
+  return $self->sqlite->db->query(
+    q{insert into minion_jobs
+       (args, attempts, delayed, expires, lax, notes, parents, priority, queue, task)
+      values (?, ?, datetime('now', ? || ' seconds'),
+      case when ? is not null then datetime('now', ? || ' seconds') end,
+      ?, ?, ?, ?, ?, ?)},
+    {json => $args}, $options->{attempts} // 1, $options->{delay} // 0,
+    @$options{qw(expire expire)}, $options->{lax} ? 1 : 0, {json => $options->{notes} || {}},
+    {json => ($options->{parents} || [])}, $options->{priority} // 0,
+    $options->{queue} // 'default', $task
+  )->last_insert_id;
 }
 
 sub fail_job   { shift->_update(1, @_) }
@@ -122,11 +121,6 @@ sub list_jobs {
     push @where, @$queues ? "queue in ($queues_in)" : 'queue is null';
     push @where_params, @$queues;
   }
-  if (defined(my $sequences = $options->{sequences})) {
-    my $sequences_in = join ',', ('?')x@$sequences;
-    push @where, @$sequences ? "sequence in ($sequences_in)" : 'sequence is null';
-    push @where_params, @$sequences;
-  }
   if (defined(my $states = $options->{states})) {
     my $states_in = join ',', ('?')x@$states;
     push @where, @$states ? "state in ($states_in)" : 'state is null';
@@ -137,6 +131,7 @@ sub list_jobs {
     push @where, @$tasks ? "task in ($tasks_in)" : 'task is null';
     push @where_params, @$tasks;
   }
+  push @where, q{(state != 'inactive' or expires is null or expires > datetime('now'))};
 
   my $where_str = @where ? 'where ' . join(' and ', @where) : '';
 
@@ -146,10 +141,10 @@ sub list_jobs {
          from minion_jobs as child, json_each(child.parents) as parent_id
          where j.id = parent_id.value) as children,
        strftime('%s',created) as created, strftime('%s',delayed) as delayed,
-       strftime('%s',finished) as finished, next, notes, parents,
-       (select id from minion_jobs where sequence = j.sequence and next = j.id) as previous,
-       priority, queue, result, strftime('%s',retried) as retried, retries,
-       sequence, strftime('%s',started) as started, state, task,
+       strftime('%s',expires) as expires, strftime('%s',finished) as finished,
+       lax, notes, parents, priority, queue, result,
+       strftime('%s',retried) as retried, retries,
+       strftime('%s',started) as started, state, task,
        strftime('%s','now') as time, worker
        from minion_jobs as j
        $where_str order by id desc limit ? offset ?},
@@ -302,6 +297,15 @@ sub repair {
       where notified < datetime('now', '-' || ? || ' seconds')}, $minion->missing_after
   );
 
+  # Old jobs with no unresolved dependencies and expired jobs
+  $db->query(
+    q{delete from minion_jobs
+      where (finished <= datetime('now', '-' || ? || ' seconds')
+      and state = 'finished' and id not in (select distinct parent_id.value
+        from minion_jobs as child, json_each(child.parents) as parent_id
+        where child.state <> 'finished'))
+      or (expires <= datetime('now') and state = 'inactive')}, $minion->remove_after);
+
   # Jobs with missing worker (can be retried)
   my $fail = $db->query(
     q{select id, retries from minion_jobs as j
@@ -315,14 +319,6 @@ sub repair {
     q{update minion_jobs set state = 'failed', result = json_quote('Job appears stuck in queue')
       where state = 'inactive' and delayed < datetime('now', '-' || ? || ' seconds')},
     $minion->stuck_after);
-
-  # Old jobs with no unresolved dependencies
-  $db->query(
-    q{delete from minion_jobs
-      where finished <= datetime('now', '-' || ? || ' seconds')
-      and state = 'finished' and id not in (select distinct parent_id.value
-        from minion_jobs as child, json_each(child.parents) as parent_id
-        where child.state <> 'finished')}, $minion->remove_after);
 }
 
 sub reset {
@@ -349,13 +345,15 @@ sub retry_job {
   return !!$self->sqlite->db->query(
     q{update minion_jobs
       set attempts = coalesce(?, attempts),
-        delayed = (datetime('now', ? || ' seconds')),
-        parents = coalesce(?, parents), priority = coalesce(?, priority),
+        delayed = datetime('now', ? || ' seconds'),
+        expires = case when ? is not null then datetime('now', ? || ' seconds') else expires end,
+        lax = coalesce(?, lax), parents = coalesce(?, parents), priority = coalesce(?, priority),
         queue = coalesce(?, queue), retried = datetime('now'),
         retries = retries + 1, state = 'inactive'
       where id = ? and retries = ?},
-    $options->{attempts}, $options->{delay} // 0, $parents,
-    @$options{qw(priority queue)}, $id, $retries
+    $options->{attempts}, $options->{delay} // 0, @$options{qw(expire expire)},
+    exists $options->{lax} ? $options->{lax} ? 1 : 0 : undef,
+    $parents, @$options{qw(priority queue)}, $id, $retries
   )->rows;
 }
 
@@ -363,7 +361,8 @@ sub stats {
   my $self = shift;
 
   my $stats = $self->sqlite->db->query(
-    q{select count(case state when 'inactive' then 1 end) as inactive_jobs,
+    q{select count(case when state = 'inactive' and (expires is null or expires > datetime('now'))
+        then 1 end) as inactive_jobs,
       count(case state when 'active' then 1 end) as active_jobs,
       count(case state when 'failed' then 1 end) as failed_jobs,
       count(case state when 'finished' then 1 end) as finished_jobs,
@@ -396,20 +395,6 @@ sub unregister_worker {
   shift->sqlite->db->query('delete from minion_workers where id = ?', shift);
 }
 
-sub _enqueue {
-  my ($db, $task, $args, $options) = @_;
-
-  return $db->query(
-    q{insert into minion_jobs
-       (args, attempts, delayed, notes, parents, priority, queue, sequence, task)
-      values (?, ?, (datetime('now', ? || ' seconds')), ?, ?, ?, ?, ?, ?)},
-    {json => $args}, $options->{attempts} // 1,
-    $options->{delay} // 0, {json => $options->{notes} || {}},
-    {json => ($options->{parents} || [])}, $options->{priority} // 0,
-    $options->{queue} // 'default', $options->{sequence}, $task
-  )->last_insert_id;
-}
-
 sub _try {
   my ($self, $id, $options) = @_;
 
@@ -426,10 +411,13 @@ sub _try {
        where delayed <= datetime('now') and id = coalesce(?, id)
        and (json_array_length(parents) = 0 or not exists (
          select 1 from minion_jobs as parent, json_each(j.parents) as parent_id
-         where parent.id = parent_id.value
-         and parent.state in ('inactive', 'active', 'failed')
-       )) and queue in ($queues_in) and state = 'inactive'
-       and task in ($tasks_in)
+         where parent.id = parent_id.value and case parent.state
+           when 'active' then 1
+           when 'failed' then not j.lax
+           when 'inactive' then (parent.expires is null or parent.expires > datetime('now'))
+         end
+       )) and queue in ($queues_in) and state = 'inactive' and task in ($tasks_in)
+       and (expires is null or expires > datetime('now'))
        order by priority desc, id
        limit 1}, $options->{id}, @$queues, @$tasks
   );
@@ -460,10 +448,7 @@ sub _update {
   )->rows > 0;
   
   my $row = $db->query('select attempts from minion_jobs where id = ?', $id)->array;
-  return 1 if !$fail || (my $attempts = $row->[0]) == 1;
-  return 1 if $retries >= ($attempts - 1);
-  my $delay = $self->minion->backoff->($retries);
-  return $self->retry_job($id, $retries, {delay => $delay});
+  return $fail ? $self->auto_retry_job($id, $retries, $row->[0]) : 1;
 }
 
 1;
@@ -625,6 +610,21 @@ L<Minion/"backoff"> after the first attempt, defaults to C<1>.
 
 Delay job for this many seconds (from now).
 
+=item expire
+
+  expire => 300
+
+Job is valid for this many seconds (from now) before it expires. Note that this
+option is B<EXPERIMENTAL> and might change without warning!
+
+=item lax
+
+  lax => 1
+
+Existing jobs this job depends on may also have transitioned to the C<failed>
+state to allow for it to be processed, defaults to C<false>. Note that this
+option is B<EXPERIMENTAL> and might change without warning!
+
 =item notes
 
   notes => {foo => 'bar', baz => [1, 2, 3]}
@@ -650,14 +650,6 @@ Job priority, defaults to C<0>. Jobs with a higher priority get performed first.
 
 Queue to put job in, defaults to C<default>.
 
-=item sequence
-
-  sequence => 'host:mojolicious.org'
-
-Sequence this job belongs to. The previous job from the sequence will be
-automatically added as a parent to continue the sequence. Note that this option
-is B<EXPERIMENTAL> and might change without warning!
-
 =back
 
 =head2 fail_job
@@ -737,13 +729,6 @@ List only jobs with these ids.
 
 List only jobs in these queues.
 
-=item sequences
-
-  sequences => ['host:localhost', 'host:mojolicious.org']
-
-List only jobs from these sequences. Note that this option is B<EXPERIMENTAL>
-and might change without warning!
-
 =item states
 
   states => ['inactive', 'active']
@@ -792,6 +777,12 @@ Epoch time job was created.
 
 Epoch time job was delayed to.
 
+=item expires
+
+  expires => 784111777
+
+Epoch time job is valid until before it expires.
+
 =item finished
 
   finished => 784111777
@@ -804,11 +795,12 @@ Epoch time job was finished.
 
 Job id.
 
-=item next
+=item lax
 
-  next => 10024
+  lax => 0
 
-Next job in sequence.
+Existing jobs this job depends on may also have failed to allow for it to be
+processed.
 
 =item notes
 
@@ -822,12 +814,6 @@ Hash reference with arbitrary metadata for this job.
 
 Jobs this job depends on.
 
-=item previous
-
-  previous => 10022
-
-Previous job in sequence.
-
 =item priority
 
   priority => 3
@@ -858,12 +844,6 @@ Epoch time job has been retried.
 
 Number of times job has been retried.
 
-=item sequence
-
-  sequence => 'host:mojolicious.org'
-
-Sequence name.
-
 =item started
 
   started => 784111777
@@ -1137,6 +1117,21 @@ Number of times performing this job will be attempted.
 
 Delay job for this many seconds (from now).
 
+=item expire
+
+  expire => 300
+
+Job is valid for this many seconds (from now) before it expires. Note that this
+option is B<EXPERIMENTAL> and might change without warning!
+
+=item lax
+
+  lax => 1
+
+Existing jobs this job depends on may also have transitioned to the C<failed>
+state to allow for it to be processed, defaults to C<false>. Note that this
+option is B<EXPERIMENTAL> and might change without warning!
+
 =item parents
 
   parents => [$id1, $id2, $id3]
@@ -1357,3 +1352,36 @@ alter table minion_jobs add column sequence text;
 alter table minion_jobs add column next integer;
 create unique index minion_jobs_next on minion_jobs (next);
 create index minion_jobs_sequence_next on minion_jobs (sequence, next);
+
+-- 10 up
+create table minion_jobs_NEW (
+  id       integer not null primary key autoincrement,
+  args     text not null,
+  created  text not null default current_timestamp,
+  delayed  text not null,
+  finished text,
+  priority integer not null,
+  result   text,
+  retried  text,
+  retries  integer not null default 0,
+  started  text,
+  state    text not null default 'inactive',
+  task     text not null,
+  worker   integer,
+  queue    text not null default 'default',
+  attempts integer not null default 1,
+  parents  text not null default '[]',
+  notes    text not null
+    check(json_valid(notes) and json_type(notes) = 'object') default '{}',
+  expires  text,
+  lax      boolean not null default 0
+);
+insert into minion_jobs_NEW
+  (id,args,created,delayed,finished,priority,result,retried,retries,
+  started,state,task,worker,queue,attempts,parents,notes)
+  select id,args,created,delayed,finished,priority,result,retried,retries,
+  started,state,task,worker,queue,attempts,parents,notes from minion_jobs;
+drop table minion_jobs;
+alter table minion_jobs_NEW rename to minion_jobs;
+create index if not exists minion_jobs_state_priority_id on minion_jobs (state, priority desc, id);
+create index minion_jobs_expires on minion_jobs (expires);
